/**************************************************************************
* DSemu: Memory management and handling (mmu.c)                           *
* Released under the terms of the BSD Public Licence                      *
* Imran Nazar (tf@oopsilon.com), 2004                                     *
**************************************************************************/

#include <stdio.h>
#include "vtbl.h"
#include "err.h"
#include "mmu.h"
#include "dma.h"
#include "ioreg.h"
#include "gpu.h"

extern u16 *VRAM;
extern u8  *VRAM8;

extern u32 LCDcolour[32768];
extern u16 BGPAL[256];
extern u16 OBJPAL[256];
extern u8 *BGPAL8;
extern u8 *OBJPAL8;

RAMWORD *MMUbanks[16]={NULL};
u32 MMUmask[16]={0x1FF,0,0x3FFFF,0x7FFF,0,0,0,0,
                 0xFFFFFF,0xFFFFFF,0xFFFFFF,
                 0xFFFFFF,0xFFFFFF,0xFFFFFF,0,0};

int MMUinit(char *file)
{
    FILE *fp; u8 c; int a;

    ROMfile=file;

    EWRAM=(RAMWORD*)malloc(65536*sizeof(RAMWORD));
    if(!EWRAM) RETFAIL("MMU: FAIL: EWRAM allocation.");

    IWRAM=(RAMWORD*)malloc(8192*sizeof(RAMWORD));
    if(!IWRAM) RETFAIL("MMU: FAIL: IWRAM allocation.");

    fp=fopen(file,"rb");
    if(!fp) RETFAIL("MMU: FAIL: ROM file open.");
    logvt->append("MMU: ROM file open.");

    fseek(fp,0,SEEK_END); ROMsize=ftell(fp); fseek(fp,0,SEEK_SET);
    GameROM=(RAMWORD*)malloc((ROMsize/4)*sizeof(RAMWORD));
    if(!GameROM) RETFAIL("MMU: FAIL: GameROM allocation.");

    if(MMUreset()) RETFAIL("MMU: FAIL: Reset.");

    MMUbanks[0x0]=BIOS;    MMUbanks[0x2]=EWRAM;
    MMUbanks[0x3]=IWRAM;
    MMUbanks[0x8]=GameROM; MMUbanks[0x9]=GameROM;
    MMUbanks[0xA]=GameROM; MMUbanks[0xB]=GameROM;
    MMUbanks[0xC]=GameROM; MMUbanks[0xD]=GameROM;
    logvt->append("MMU: Game ROM loaded.");

    RETPASS("MMU: Initialised.");
}

int MMUreset()
{
    FILE *fp; u8 c; int a; int BIOSsize;

    memset(EWRAM,0,0x3FFFF);
    memset(IWRAM,0,0x7FFF);

    fp=fopen(ROMfile,"rb");
    if(!fp) RETFAIL("MMU: FAIL: ROM file open.");

    for(a=0;a<ROMsize;a++)
    {
    	c=fgetc(fp);
        #if WORDS_BIGENDIAN
        GameROM[a>>2].b[3-(a&3)]=c;
        #else
        GameROM[a>>2].b[a&3]=c;
        #endif
    }
    for(a=0;a<(ROMsize>>2);a++)
    {
    	GameROM[a].op=ARM7opUNL;
        GameROM[a].cond=0x000E;
        GameROM[a].cyc=1;
    }
    fclose(fp);

    BIOS=(RAMWORD*)malloc(0x80*sizeof(RAMWORD));
    if(!BIOS) RETFAIL("MMU: FAIL: BIOS allocation.");

    fp=fopen("bioshack.bin","rb");
    if(!fp) RETFAIL("MMU: FAIL: BIOS file open.");

    for(a=0;a<0x200;a++)
    {
    	c=fgetc(fp);
        #if WORDS_BIGENDIAN
        BIOS[a>>2].b[3-(a&3)]=c;
        #else
        BIOS[a>>2].b[a&3]=c;
        #endif
    }
    for(a=0;a<0x80;a++)
    {
    	BIOS[a].op=ARM7opUNL;
        BIOS[a].cond=0x000E;
        BIOS[a].cyc=1;
    }
    fclose(fp);
}

void MMUfini()
{
    free(EWRAM);
    free(IWRAM);
    free(GameROM);
    logvt->append("MMU: Shutdown.");
}

#if WORDS_BIGENDIAN

#define MMUintRdB(ram,addr) b=ram[addr].b[3-(addr&3)]
#define MMUintRdH(ram,addr) \
    h=(addr&2)? \
      ((ram[addr].b[0])| \
       (ram[addr].b[1]<<8)): \
      ((ram[addr].b[2])| \
       (ram[addr].b[3]<<8))
#define MMUintRdW(ram,addr) \
    w=(ram[addr].b[0])| \
      (ram[addr].b[1]<<8)| \
      (ram[addr].b[2]<<16)| \
      (ram[addr].b[3]<<24)

#else

#define MMUintRdB(ram,addr) b=ram[(addr)>>2].b[(addr)&3]
#define MMUintRdH(ram,addr) h=ram[(addr)>>2].h[((addr)&2)>>1]
#define MMUintRdW(ram,addr) w=ram[(addr)>>2].data

#endif

u8 MMUrdB(u32 prot, u32 addr)
{
    u8 b;
    switch(addr&0x0F000000)
    {
        case 0x00000000:
            MMUintRdB(BIOS, addr&0x000001FF);
            break;
        case 0x02000000:
	    MMUintRdB(EWRAM, addr&0x0003FFFF);
	    ARM7addClock(2); break;
	case 0x03000000:
	    MMUintRdB(IWRAM, addr&0x00007FFF);
	    break;
	case 0x04000000:
	    if(prot || GBAio[(addr&0x3FF)>>1].flags&REG_FLAG_R)
                b=GBAio[(addr&0x3FF)>>1].b[addr&1];
            else b=0;
	    break;
	case 0x05000000:
            if(addr&0x00000200)
	        b=(addr&1)?OBJPAL[(addr&0x1FF)>>1]>>8:OBJPAL[(addr&0x1FF)>>1]&255;
            else
	        b=(addr&1)?BGPAL[(addr&0x1FF)>>1]>>8:BGPAL[(addr&0x1FF)>>1]&255;
            break;
	case 0x06000000:
	    b=(addr&1)?VRAM[(addr&0x0001FFFF)>>1]>>8:VRAM[(addr&0x0001FFFF)>>1]&255;
	    break;
	case 0x08000000: case 0x09000000:
	    MMUintRdB(GameROM, addr&0x00FFFFFF);
	    ARM7addClock(4); break;
	case 0x0A000000: case 0x0B000000:
	    MMUintRdB(GameROM, addr&0x00FFFFFF);
	    ARM7addClock(5); break;
	case 0x0C000000: case 0x0D000000:
	    MMUintRdB(GameROM, addr&0x00FFFFFF);
	    ARM7addClock(6); break;
	default: break;
    }
    return b;
}

u16 MMUrdH(u32 prot, u32 addr)
{
    u16 h;
    switch(addr&0x0F000000)
    {
        case 0x00000000:
            MMUintRdH(BIOS, addr&0x000001FF);
            break;
        case 0x02000000:
	    MMUintRdH(EWRAM, addr&0x0003FFFF);
	    ARM7addClock(2); break;
	case 0x03000000:
	    MMUintRdH(IWRAM, addr&0x00007FFF);
	    break;
	case 0x04000000:
	    if(prot || GBAio[(addr&0x3FF)>>1].flags&REG_FLAG_R)
                h=GBAio[(addr&0x3FF)>>1].data;
            else h=0;
	    break;
	case 0x05000000:
            if(addr&0x00000200) h=OBJPAL[(addr&0x1FF)>>1];
            else                h=BGPAL [(addr&0x1FF)>>1];
            break;
	case 0x06000000:
	    h=VRAM[(addr&0x0001FFFF)>>1];
	    break;
	case 0x08000000: case 0x09000000:
	    MMUintRdH(GameROM, addr&0x00FFFFFF);
	    ARM7addClock(4); break;
	case 0x0A000000: case 0x0B000000:
	    MMUintRdH(GameROM, addr&0x00FFFFFF);
	    ARM7addClock(5); break;
	case 0x0C000000: case 0x0D000000:
	    MMUintRdH(GameROM, addr&0x00FFFFFF);
	    ARM7addClock(6); break;
	default: break;
    }
    return h;
}

u32 MMUrdW(u32 prot, u32 addr)
{
    u32 w; u16 hl,hh;
    switch(addr&0x0F000000)
    {
        case 0x00000000:
            MMUintRdW(BIOS, addr&0x000001FF);
            break;
        case 0x02000000:
	    MMUintRdW(EWRAM, addr&0x0003FFFF);
	    ARM7addClock(5); break;
	case 0x03000000:
	    MMUintRdW(IWRAM, addr&0x00007FFF);
	    break;
	case 0x04000000:
	    if(prot || GBAio[(addr&0x3FF)>>1].flags&REG_FLAG_R)
                hl=GBAio[(addr&0x3FF)>>1].data;
            else hl=0;
	    addr+=2;
	    if(prot || GBAio[(addr&0x3FF)>>1].flags&REG_FLAG_R)
                hh=GBAio[(addr&0x3FF)>>1].data;
            else hh=0;
            w=hl+(hh<<16);
	    break;
	case 0x05000000:
            if(addr&0x00000200) hl=OBJPAL[(addr&0x1FF)>>1];
            else                hl=BGPAL [(addr&0x1FF)>>1];
            addr+=2;
            if(addr&0x00000200) hh=OBJPAL[(addr&0x1FF)>>1];
            else                hh=BGPAL [(addr&0x1FF)>>1];
            w=hl+(hh<<16);
            break;
	case 0x06000000:
	    hl=VRAM[(addr&0x0001FFFF)>>1]; addr+=2;
	    hh=VRAM[(addr&0x0001FFFF)>>1];
            w=hl+(hh<<16);
            break;
	case 0x08000000: case 0x09000000:
	    MMUintRdW(GameROM, addr&0x00FFFFFF);
	    ARM7addClock(7); break;
	case 0x0A000000: case 0x0B000000:
	    MMUintRdW(GameROM, addr&0x00FFFFFF);
	    ARM7addClock(8); break;
	case 0x0C000000: case 0x0D000000:
	    MMUintRdW(GameROM, addr&0x00FFFFFF);
	    ARM7addClock(9); break;
	default: break;
    }
    return w;
}

RAMWORD MMUrdS(u32 prot, u32 addr)
{
/*    RAMWORD def;
    def.data=0xEC000000;*/
    switch(addr&0x0F000000)
    {
        case 0x00000000: return BIOS[(addr&0x001FF)>>2]; break;
        case 0x02000000: return EWRAM[(addr&0x3FFFF)>>2]; break;
        case 0x03000000: return IWRAM[(addr&0x07FFF)>>2]; break;
	case 0x08000000:
	case 0x09000000:
	case 0x0A000000:
	case 0x0B000000:
	case 0x0C000000:
        case 0x0D000000: return GameROM[(addr&0xFFFFFF)>>2]; break;
	//default: return def; break;
    }
/*    int bank=(addr>>24)&15;
    if(MMUbanks[bank]!=NULL)
        return MMUbanks[bank][(addr&MMUmask[bank])>>2];*/

}

#if WORDS_BIGENDIAN

#define MMUintWrB(ram,addr,val) \
    ram[addr].b[3-(addr&3)]=val; \
    ram[addr].op=ARM7opUNL; \
    ram[addr].cond=0x000E; \
    ram[addr].cyc=1
#define MMUintWrH(ram,addr,val) \
    if(addr&2){ \
        ram[addr].b[0]=val&255; \
        ram[addr].b[1]=val>>8; \
    }else{ \
        ram[addr].b[2]=val&255; \
        ram[addr].b[3]=val>>8; \
    } \
    ram[addr].op=ARM7opUNL; \
    ram[addr].cond=0x000E; \
    ram[addr].cyc=1
#define MMUintWrW(ram,addr,val) \
    ram[addr].b[0]=val&255; \
    ram[addr].b[1]=val>>8; \
    ram[addr].b[2]=val>>16; \
    ram[addr].b[3]=val>>24; \
    ram[addr].op=ARM7opUNL; \
    ram[addr].cond=0x000E; \
    ram[addr].cyc=1

#else

#define MMUintWrB(ram,addr,val) \
    ram[(addr)>>2].b[(addr)&3]=val; \
    ram[(addr)>>2].op=ARM7opUNL; \
    ram[(addr)>>2].cond=0x000E; \
    ram[(addr)>>2].cyc=1
#define MMUintWrH(ram,addr,val) \
    ram[(addr)>>2].h[((addr)&2)>>1]=val; \
    ram[(addr)>>2].op=ARM7opUNL; \
    ram[(addr)>>2].cond=0x000E; \
    ram[(addr)>>2].cyc=1
#define MMUintWrW(ram,addr,val) \
    ram[(addr)>>2].data=val; \
    ram[(addr)>>2].op=ARM7opUNL; \
    ram[(addr)>>2].cond=0x000E; \
    ram[(addr)>>2].cyc=1

#endif

void MMUwrB(u32 bus, u32 addr, u8 val)
{
    switch(addr&0x0F000000)
    {
        case 0x02000000:
	    MMUintWrB(EWRAM, addr&0x0003FFFF, val);
	    ARM7addClock(2); break;
	case 0x03000000:
	    MMUintWrB(IWRAM, addr&0x00007FFF, val);
	    break;
	default: break;
    }
}

void MMUwrH(u32 bus, u32 addr, u16 val)
{
    switch(addr&0x0F000000)
    {
        case 0x02000000:
	    MMUintWrH(EWRAM, addr&0x0003FFFF, val);
	    ARM7addClock(2); break;
	case 0x03000000:
	    MMUintWrH(IWRAM, addr&0x00007FFF, val);
	    break;
	case 0x04000000:
            if(GBAio[(addr&0x3FF)>>1].flags&REG_FLAG_W)
            {
                if((addr&0x3FE)==0x202) IntClear(val);
                else GBAio[(addr&0x3FF)>>1].data=val;
                DMAcheck(DMA_TIMENOW);
            }
	    break;
	case 0x05000000:
	    if(addr&0x00000200) OBJPAL[(addr&0x1FF)>>1]=val;
	    else                BGPAL [(addr&0x1FF)>>1]=val;
	    break;
	case 0x06000000:
	    VRAM[(addr&0x0001FFFF)>>1]=val; break;
	default: break;
    }
}

void MMUwrW(u32 bus, u32 addr, u32 val)
{
    switch(addr&0x0F000000)
    {
        case 0x02000000:
	    MMUintWrW(EWRAM, addr&0x0003FFFF, val);
	    ARM7addClock(2); break;
	case 0x03000000:
	    MMUintWrW(IWRAM, addr&0x00007FFF, val);
	    break;
	case 0x04000000:
            if(GBAio[(addr&0x3FF)>>1].flags&REG_FLAG_W)
            {
                if((addr&0x3FE)==0x202) IntClear(val&65535);
                else GBAio[(addr&0x3FF)>>1].data=val&65535;
            }
	    addr+=2;
            if(GBAio[(addr&0x3FF)>>1].flags&REG_FLAG_W)
            {
                if((addr&0x3FE)==0x202) IntClear(val>>16);
                else GBAio[(addr&0x3FF)>>1].data=val>>16;
            }
            DMAcheck(DMA_TIMENOW);
	    break;
	case 0x05000000:
	    if(addr&0x00000200) OBJPAL[(addr&0x1FF)>>1]=val&65535;
	    else                BGPAL [(addr&0x1FF)>>1]=val&65535;
	    addr+=2;
	    if(addr&0x00000200) OBJPAL[(addr&0x1FF)>>1]=val>>16;
	    else                BGPAL [(addr&0x1FF)>>1]=val>>16;
	case 0x06000000:
	    VRAM[(addr&0x0001FFFF)>>1]=val&65535; addr+=2;
	    VRAM[(addr&0x0001FFFF)>>1]=val>>16;
	    break;
	default: break;
    }
}

void MMUwrS(u32 bus, u32 addr, RAMWORD val)
{
    switch(addr&0x0F000000)
    {
        case 0x02000000: EWRAM[(addr&0x3FFFF)>>2]=val; break;
        case 0x03000000: IWRAM[(addr&0x07FFF)>>2]=val; break;
	case 0x08000000:
	case 0x09000000:
	case 0x0A000000:
	case 0x0B000000:
	case 0x0C000000:
        case 0x0D000000: GameROM[(addr&0xFFFFFF)>>2]=val; break;
	default: break;
    }
}

/*** EOF:mmu.c ***********************************************************/

